import type {configDependency} from '../core/Config.js'
import type {Dependencies, OpType} from '../interfaces/type.js'

export const add:        OpType<'add',        number> = (a, b) => a + b
export const unaryMinus: OpType<'unaryMinus', number> = a => -a
export const conj:       OpType<'conj',       number> = a => a
export const subtract:   OpType<'subtract',   number> = (a, b) => a - b
export const multiply:   OpType<'multiply',   number> = (a, b) => a * b
export const absquare:   OpType<'absquare',   number> = a => a * a
export const reciprocal: OpType<'reciprocal', number> = a => 1 / a
export const divide:     OpType<'divide',     number> = (a, b) => a / b

const basicSqrt = a => isNaN(a) ? NaN : Math.sqrt(a)
export const conservativeSqrt: OpType<'conservativeSqrt', number> = basicSqrt

export const sqrt =
   (dep: configDependency & Dependencies<'complex', number>):
   OpType<'sqrt', number> => {
      if (dep.config.predictable || !dep.complex) return basicSqrt
      return a => {
         if (isNaN(a)) return NaN
         if (a >= 0) return Math.sqrt(a)
         return dep.complex(0, Math.sqrt(unaryMinus(a)))
      }
   }
